/*	$Id: inode.c,v 1.34 2005/09/01 21:10:16 phil Exp $ */

/* BLURB gpl

                           Coda File System
                              Release 6

          Copyright (c) 1987-2005 Carnegie Mellon University
                  Additional copyrights listed below

This  code  is  distributed "AS IS" without warranty of any kind under
the terms of the GNU General Public Licence Version 2, as shown in the
file  LICENSE.  The  technical and financial  contributors to Coda are
listed in the file CREDITS.

*/

//
// inode.c : "inodes" and "filehandles".
// 

#include <dllist.h>
#include <coda_types.h>
#include <coda.h>
#include <coda_nt.h>
#include <cnode.h>
#include <ctype.h>

// A constant to help decide if we want to free or reuse and
// inode.   Number is a guess!

#define MIN_INODES 250

// Should be in a .h file somewhere!
extern PDEVICE_OBJECT coda_device_object;

#define CACHE_SIZE 1023
#define MAX_NHASH_SHIFT 3
struct dllist_head inode_fid_cache[CACHE_SIZE+1];
struct dllist_head inode_name_cache[CACHE_SIZE+1];

static ERESOURCE cache_lock;

// list of inode not pinned by wrapper -- inserted at tail.
// reused at head.  Nodes on this list may still be in the
// cache!  "released inodes" (e.g. deleted files) are 
// inserted at head.  (if the list is long enough, released
// inodes are freed instead of being put on the free list.

struct dllist_head inode_freelist;
static int free_count; 

// list of all inodes ... so we can free all at umount time.
struct dllist_head inode_all;

// inode globals
static int inode_count;

// file handle globals
static coda_filehandle *fh_free;
static int fh_unique;


// Missing strcasecmp
int strcasecmp (char *str1, char *str2)
{
    while ((tolower(*str1) == tolower(*str2)))
	if (*str1++ == 0)
	    return (0);
	else
	    str2++;
    return (tolower(*str1) - tolower(*str2));
} 

void inode_init()
{
    int i;
    
    // The cache ...
    for ( i = 0 ; i <= CACHE_SIZE ; i++ ) { 
	list_init(&(inode_fid_cache[i]));
	list_init(&(inode_name_cache[i]));
    }

    list_init(&inode_all);
    inode_count = 0;

    list_init(&inode_freelist);
    free_count = 0;

    // File handles ...
    fh_free = NULL;
    fh_unique = 1;

    // Initialize the global resource lock
    ExInitializeResourceLite (&CodaWrapperLock);
    ExInitializeResourceLite (&cache_lock);
}

// Hash functions.

static int inode_fhash(CodaFid *fid)
{
    return (CACHE_SIZE & (coda_f2i(fid)));
}

static int inode_nhash(char *name, CodaFid *pfid) 
{
    int len = strlen(name);
    int val = inode_fhash(pfid) + len;
    int i;
    
    for (i=0; i<len; i++)
	val += tolower(name[i]) << (i & MAX_NHASH_SHIFT);

    return (CACHE_SIZE & val);
}

// Insert -- don't put on freelist if pinned or already there!

void inode_hash_insert(coda_inode *inode)
{
    int fhash = inode_fhash(&inode->c_fid);
    int nhash = inode_nhash(inode->c_name, &(inode->c_parent->c_fid));

    CDEBUG ((D_INODE,
	     "inode_hash_insert: inode %s [%s] fhash = %d, nhash = %d\n",
	     inode->c_name, coda_f2s(&inode->c_fid), fhash, nhash));

    list_add(&inode->c_fidlist,  &(inode_fid_cache[fhash]));
    list_add(&inode->c_namelist, &(inode_name_cache[nhash]));
    if ( list_empty(&inode->c_freelist) && !inode_hasflag(inode, I_PINNED)) {
	list_add(&inode->c_freelist, inode_freelist.prev);
	free_count++;
    }
}

void inode_rename(coda_inode *inode, coda_inode *parent, char *name)
{

    // Remove from cache lists
    LockCache();
    inode_cache_delete (inode);

    inode->c_parent = parent;
    strncpy (inode->c_name, name, CODA_MAXNAMLEN);

    // Reinsert
    inode_hash_insert (inode);
    UnLockCache();
}

// Utility routines

int inode_fideq(CodaFid *f1, CodaFid *f2)
{   
    return ((f1->opaque[0] == f2->opaque[0]) &&
            (f1->opaque[1] == f2->opaque[1]) &&
            (f1->opaque[2] == f2->opaque[2]) &&
            (f1->opaque[3] == f2->opaque[3]));
}

coda_inode *inode_find_fid(CodaFid *fid)
{
    int hash = inode_fhash(fid);
    struct dllist_head *lh, *le;
    coda_inode *inode;
    int max = inode_count;

    CDEBUG((D_INODE, "inode_find_fid: hash for %s is %d, max is %d\n", 
	    coda_f2s(fid), hash, max));

    LockCache();
    le = &(inode_fid_cache[hash]);
    lh = le;
    for ( le = lh->next ; (le != lh) && max ; le = le->next, max-- ) {
	inode = list_entry(le, coda_inode, c_fidlist);
	CDEBUG ((D_INODE | D_DETAILS,
		 "inode_find_fid: at %s\n", coda_f2s(&inode->c_fid)));
	if ( inode_fideq(&inode->c_fid, fid) ) {
	    UnLockCache();
	    return inode ; 
	}
    }
    UnLockCache();
    
#if DBG
    if (max == 0)
      DbgPrint("inode_find_fid: drop out at max = %d hash = %d\n", max, hash);
#endif

    return NULL;
}

coda_inode *inode_find_name(coda_inode *parent, char *name, int CaseInsensitive)
{
    int hash = inode_nhash(name, &parent->c_fid);
    struct dllist_head *lh, *le;
    coda_inode *inode;
    int max = inode_count;

    CDEBUG ((D_INODE, "inode_find_name: name = %s, parent = 0x%x,  nhash = %d\n",
	     name, parent, hash));

    LockCache();
    le = &(inode_name_cache[hash]);
    lh = le;
    for ( le = lh->next ; (le != lh) && max ; le = le->next, max-- ) {
	inode = list_entry(le, coda_inode, c_namelist);
	CDEBUG ((D_INODE | D_DETAILS,
		 "inode_find_name: at %s 0x%x (str %d, case %d)\n",
		 inode->c_name, inode->c_parent, strcmp(name, inode->c_name),
		 strcasecmp(name, inode->c_name)));
	if ( inode->c_parent != parent )
	    continue ;
	if ( (strcmp(name, inode->c_name) == 0)
	     || (CaseInsensitive && (strcasecmp(name, inode->c_name) == 0)) ) {
	    UnLockCache();
	    return inode;
	}
    }
    UnLockCache();

#if DBG
    if (max == 0)
      DbgPrint("inode_find_name: drop out at max = %d, hash = %d name = '%s'\n",
	       max, hash, name);
#endif

    return NULL;
}


// Get a new inode!

coda_inode *inode_new(CodaFid *fid, coda_inode *parent, char *name)
{
    int fidhash, namehash;
    coda_inode *inode = NULL;

    if ( !list_empty (&inode_freelist) && inode_count > MIN_INODES) {

	// Reuse an inode
	inode = list_entry(inode_freelist.next, coda_inode, c_freelist);
	
	CDEBUG ((D_INODE, "inode_new: Reusing inode 0x%x\n", inode));

	LockCache();
	inode_cache_delete (inode);
	list_del(&inode->c_freelist);
	free_count--;
	if (!inode_hasflag(inode, I_REL) && inode->c_parent != inode)
	    inode->c_parent->c_count--;
	UnLockCache();

    } else {

	// Create a new one.
    
	inode = (coda_inode *) coda_malloc(sizeof(*inode));
	if (!inode) {
	    DbgPrint("Out of memory!\n");
	    return NULL;
	}

	LockCache();
	inode_count++;

	// Initialize fields that won't be zeroed later

	inode->c_magic = CODA_MAGIC;
	list_add(&inode->c_alllist, &inode_all);
	UnLockCache();
	list_init(&inode->c_fidlist);
	list_init(&inode->c_namelist);
	list_init(&inode->c_freelist);
	ExInitializeResourceLite(&inode->c_lock);

	CDEBUG ((D_INODE, "inode_new: Made new inode 0x%x\n", inode));
    }

    // Make root node have self as parent
    if ( !parent ) 
	parent = inode;

    // Initialize all to zero
    RtlZeroMemory ((void *)&(inode->c_fh_list),
		   sizeof(*inode) -
		   ((int)(&inode->c_fh_list) - (int)(inode)));
		   

    // Fill in non-zero elements 
    inode->c_fh_list = NULL;
    inode->c_fh_cnt = 0;
    inode->c_fid = *fid;
    inode->c_count = 1;
    inode->c_ntvol = &coda_nt_volume;
    inode->c_parent = parent;
    if (inode != parent) 
	parent->c_count++;
    strncpy(inode->c_name, name, CODA_MAXNAMLEN-1);
    inode->c_name[CODA_MAXNAMLEN-1] = 0;

    //RtlCopyUnicodeString(&inode->c_unicode_name, uname);

    // Initialize the event and lock
    KeInitializeEvent (&inode->c_event, NotificationEvent, FALSE);

    LockCache();
    inode_hash_insert(inode);
    UnLockCache();
    
    return inode;
}

// Need to delete an inode from the cache.

void inode_cache_delete (coda_inode *inode)
{
    if ( !list_empty(&inode->c_fidlist) ) {
	list_del(&inode->c_fidlist);
    }
    if ( !list_empty(&inode->c_namelist) ) {
	list_del(&inode->c_namelist);
    }
}


// We are really done with this inode.  Called when an
// inode needs to be not found again ... e.g. when deleted ...

void inode_release(coda_inode *inode)
{
    CDEBUG ((D_INODE, "inode_release:\n"));
    inode->c_status = I_REL;
    if (inode->c_parent != inode)
	inode->c_parent->c_count--;

    if (free_count > MIN_INODES) {
	inode_free(inode);
	return;
    }

    // Add node to the front of the free list

    LockCache();
    inode_cache_delete (inode); 
    if ( !list_empty(&inode->c_freelist) ) {
	list_del(&inode->c_freelist);
	free_count--;
    }

    list_add(&inode->c_freelist, &inode_freelist);
    free_count++;

    UnLockCache();
}

// Inode pin and unpin ... keep pinned inodes off the free list!

void inode_pin(coda_inode *inode)
{
    CDEBUG ((D_INODE, "inode_pin:\n"));
    LockCache();
    if ( !list_empty(&inode->c_freelist) ) {
	list_del(&inode->c_freelist);
	free_count--;
    } 
    
    inode_setflag (inode, I_PINNED);
    coda_nt_volume.v_refcount++;
    UnLockCache();
}

void inode_unpin(coda_inode *inode)
{
    int rc;

    CDEBUG ((D_INODE, "inode_unpin:\n"));
    if (inode_hasflag(inode, I_DELETE)) {
	CDEBUG ((D_WRAPPER, "inode_unpin: Closing file.\n"));
	if (!(--(inode->c_open_cnt))) {
	    CDEBUG((D_FILE, "inode_unpin: Closing inode file object.")); 
	    if (inode->c_object != NULL) {
		// Close the object!
		inode_close_handle(inode);
	    } else {
		DbgPrint("inode_unpin: Closing NULL c_object, bad news\n");
	    }
	}
	rc = venus_close (&(inode->c_fid), C_O_WRITE, GetNtUID());
	// XXX check rc?
    }

    LockCache();
    list_add(&inode->c_freelist, inode_freelist.prev);
    free_count++;
    inode_unsetflag (inode, I_PINNED);
    coda_nt_volume.v_refcount--;
    UnLockCache();

    if (inode_hasflag(inode, I_PURGE)) {
	CDEBUG ((D_INODE, "inode_unpin: releaseing inode\n"));
	inode_release(inode);
    }
}

// Actually delete the inode space!

void inode_free(coda_inode *inode)
{
    CDEBUG ((D_INODE, "inode_free:\n"));
    LockCache();
    inode_cache_delete (inode); 
    if ( !list_empty(&inode->c_freelist) ) {
	list_del(&inode->c_freelist);
	free_count--;
    }
    if ( !list_empty(&inode->c_alllist) )
	list_del(&inode->c_alllist);
    inode_count --;
    UnLockCache();
    
    ExDeleteResourceLite(&inode->c_lock);
    coda_free(inode);

}

// Free all the inodes ...

void inode_freeall()
{
    struct dllist_head *lh;
    coda_inode *inode;

    CDEBUG ((D_INODE, "inode_freeall:\n"));

    LockCache();
    lh = &inode_all;
    while (lh->next != lh) {
	inode = list_entry(lh->next, coda_inode, c_alllist);
	CDEBUG ((D_INODE, " freeall: %s\n", coda_f2s(&inode->c_fid)));
	inode_free (inode);
    }
    UnLockCache();
}

void wi_close_handle ( PDEVICE_OBJECT pdev, PVOID Arg )
{
    coda_inode *inode = Arg;

    CDEBUG ((D_INODE, "wi_close_handle: inode->c_work_item = 0x%x\n",
	     inode->c_work_item));

    // Close it!
    (void) ZwClose (inode->c_handle);
    ObDereferenceObject(inode->c_object);

    // Let them go!
    KeSetEvent (&inode->c_event, EVENT_INCREMENT, FALSE);

    CDEBUG ((D_INODE, "wi_close_handle: exit\n"));
}

// Need to close the c_handle in a system thread 
void inode_close_handle(coda_inode *inode)
{
    NTSTATUS status;
    PDEVICE_OBJECT fsdDevice;
    if (IoIsSystemThread(PsGetCurrentThread())) {
        (void) ZwClose (inode->c_handle);
	ObDereferenceObject(inode->c_object);
    } else {
	// Must call on a worker thread!
	CDEBUG ((D_INODE, "inode_close_handle: Sending out work\n"));
 	inode->c_work_item = IoAllocateWorkItem(coda_device_object);
	ASSERT(inode->c_work_item != NULL);  // Very Bad!  
	IoQueueWorkItem (inode->c_work_item, wi_close_handle,
			 DelayedWorkQueue, (PVOID) inode);
	status = KeWaitForSingleObject (&inode->c_event, Executive,
					KernelMode, FALSE, NULL);
	ASSERT(status == STATUS_SUCCESS);   // Very Bad!

	// Free the work item and clear event
	IoFreeWorkItem (inode->c_work_item);
	KeClearEvent(&inode->c_event);
	CDEBUG ((D_INODE, "inode_close_handle: Work done\n"));
    }

    inode->c_handle = NULL;
    inode->c_object = NULL;
}


//
// File handle routines 
//

// Get a file handle for the unique dir/inode/uid.
// uid looked up here!
// These are stored in a list in the inode!

coda_filehandle *
get_cfh (coda_inode *inode, coda_inode *dir)
{
    coda_filehandle *lfh;
    cuid_t tuid = GetNtUID();

    ENTRY (D_FHAND, "get_cfh");

    // Look for existing one
    lfh = inode->c_fh_list;
    while (lfh != NULL) {
	if (lfh->fh_dir == dir && lfh->fh_uid == tuid) {
	    CDEBUG ((D_FHAND, "get_cfh: Old fh=%d, inode=%s, fh_count=%d, "
		     "fh_uid \n",
		     lfh->fh_unique, coda_f2s(&inode->c_fid),
		     inode->c_fh_cnt, lfh->fh_uid));
	    return lfh;
	}
	lfh = lfh->fh_next;
    }

    if (fh_free == NULL) {
	lfh = (coda_filehandle *) coda_malloc (sizeof(coda_filehandle));
	if (!lfh) {
	    DbgPrint ("get_cfh: out of memory!\n");
	    return NULL;
	}
    } else {
	lfh = fh_free;
	fh_free = fh_free->fh_next;
    }

    // initialize

    lfh->fh_unique = fh_unique++;
    lfh->fh_inode  = inode;
    lfh->fh_dir    = dir;
    lfh->fh_isopen = 0;
    lfh->fh_granted = 0;
    lfh->fh_openflags = 0;
    lfh->fh_uid = tuid;
    
    // Inode fields ...
    lfh->fh_next = inode->c_fh_list;
    inode->c_fh_list = lfh;
    inode->c_fh_cnt++;
    inode_pin (inode);

    CDEBUG ((D_FHAND, "get_cfh: New fh=%d, inode=%s, fh_count=%d\n",
	     lfh->fh_unique, coda_f2s(&inode->c_fid), inode->c_fh_cnt));

    return lfh;
}

void
free_cfh (coda_filehandle *ofh)
{
    coda_filehandle *tfh;
    coda_inode      *inode;

    // Unlink from inode list
    inode = ofh->fh_inode;

    CDEBUG ((D_FHAND, "free_cfh: fh=%d, inode=%s, fh_count=%d\n",
	     ofh->fh_unique, coda_f2s(&inode->c_fid), inode->c_fh_cnt));

    if (inode->c_fh_list == ofh) {
	inode->c_fh_list = ofh->fh_next;
    } else {
	tfh = inode->c_fh_list;
	while (tfh->fh_next && tfh->fh_next != ofh)
	    tfh = tfh->fh_next;
	if (tfh && tfh->fh_next) {
	    tfh->fh_next = ofh->fh_next;
	} else {
	    DbgPrint ("free_cfh: BIG PROBLEM IN NEXT LIST!\n");
	}
    }

    // Decrement count
    inode->c_fh_cnt--;

    // Put on free list
    ofh->fh_next = fh_free;
    fh_free = ofh;

    // Unpin the inode if we are last fh.
    if (!inode->c_fh_list)
	inode_unpin(inode);

    EXIT (D_FHAND, "free_cfh");
}
    

#ifdef DBG

// Dump Cache!

void print_inode_cache()
{
    struct dllist_head *le, *lh;
    coda_inode *inode;
    int ix;

    CDEBUG ((D_PRCACHE, "print_inode_cache: *****************************\n"));
    CDEBUG ((D_PRCACHE, "  inode count = %d\n", inode_count));
    CDEBUG ((D_PRCACHE, "  free_count = %d\n", free_count));
    if (DebugLevel & D_DETAILS) {
      CDEBUG ((D_PRCACHE, "  Free List:\n"));
      lh = &inode_freelist;
      for ( le = lh->next; le != lh; le = le->next) {
	inode = list_entry(le, coda_inode, c_freelist);
	CDEBUG ((D_PRCACHE, "     %s %d%s%s%s\n", coda_f2s(&inode->c_fid),
		 inode->c_fh_cnt,
		 (inode_hasflag(inode, I_REL)? " R": ""),
		 (inode_hasflag(inode, I_PINNED)? " P": ""),
		 (inode_hasflag(inode, I_VATTR)? " V": "")));
      }
      CDEBUG ((D_PRCACHE, "  FID cache:\n"));
      for (ix = 0; ix < CACHE_SIZE; ix++) {
	if (!list_empty(&(inode_fid_cache[ix]))) {
	  CDEBUG ((D_PRCACHE, "     %d:\n", ix));
	  lh = &(inode_fid_cache[ix]);
	  for ( le = lh->next ; le != lh; le = le->next) {
	    inode = list_entry(le, coda_inode, c_fidlist);
	    CDEBUG ((D_PRCACHE, "     %s %d%s%s%s\n", coda_f2s(&inode->c_fid),
		     inode->c_fh_cnt,
		     (inode_hasflag(inode, I_REL)? " R": ""),
		     (inode_hasflag(inode, I_PINNED)? " P": ""),
		     (inode_hasflag(inode, I_VATTR)? " V": "")));
	  }
	}
      }
      
      CDEBUG ((D_PRCACHE, "  Name cache:\n"));
      for (ix = 0; ix < CACHE_SIZE; ix++) {
	if (!list_empty(&(inode_name_cache[ix]))) {
	  CDEBUG ((D_PRCACHE, "     %d:\n", ix));
	  lh = &(inode_name_cache[ix]);
	  for ( le = lh->next ; le != lh; le = le->next) {
	    inode = list_entry(le, coda_inode, c_namelist);
	    CDEBUG ((D_PRCACHE, "     %s %d%s%s%s\n", coda_f2s(&inode->c_fid),
		     inode->c_fh_cnt,
		     (inode_hasflag(inode, I_REL)? " R": ""),
		     (inode_hasflag(inode, I_PINNED)? " P": ""),
		     (inode_hasflag(inode, I_VATTR)? " V": "")));
	  }
	}
      }
      
      CDEBUG ((D_PRCACHE, "  alllist:\n"));
      lh = &inode_all;
      for ( le = lh->next; le != lh; le = le->next) {
	inode = list_entry(le, coda_inode, c_alllist);
	CDEBUG ((D_PRCACHE, "     %s %d%s%s%s\n", coda_f2s(&inode->c_fid),
		 inode->c_fh_cnt,
		 (inode_hasflag(inode, I_REL)? " R": ""),
		 (inode_hasflag(inode, I_PINNED)? " P": ""),
		 (inode_hasflag(inode, I_VATTR)? " V": "")));
      }
    }
    
    CDEBUG ((D_PRCACHE, "print_inode_cache: END  ************************\n"));
    
}
#endif
